import torch
import hydra
from ..base.constants import MODEL_MOPOEVAE
from ..base.base_model import BaseModelVAE
from itertools import combinations
from ..base.representations import ProductOfExperts, MixtureOfExperts
import numpy as np
class MoPoEVAE(BaseModelVAE):
    r"""
    Mixture-of-Product-of-Experts Variational Autoencoder.

    Code is based on: https://github.com/thomassutter/MoPoE

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.beta (int, float): KL divergence weighting term.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
    
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Sutter, Thomas & Daunhawer, Imant & Vogt, Julia. (2021). Generalized Multimodal ELBO.
    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_MOPOEVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)
        self.subsets = self.set_subsets()

        if self.weight_ll:
            self.ll_weighting = 1/self.n_views
        else:
            self.ll_weighting = 1
        
    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): list containing the MoE joint encoding distribution. If training, the model also returns the encoding distribution for each subset. 
        """
        mu = []
        logvar = []

        for i in range(self.n_views):
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_) 
            logvar.append(logvar_) 
        mu = torch.stack(mu)
        logvar = torch.stack(logvar)

        mu_out = []
        logvar_out = []
        if self._training:
            qz_xs = []
            for subset in self.subsets:
                mu_s = mu[subset]
                logvar_s = logvar[subset]
                if len(subset) == self.n_views:
                    mu_ = self.prior.mean
                    mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
                    logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
                    logvar_ = logvar_.expand(logvar[0].shape)
                    mu_ = mu_.unsqueeze(0)
                    logvar_ = logvar_.unsqueeze(0)
                    mu_s = torch.cat([mu_s, mu_], dim=0)
                    logvar_s = torch.cat([logvar_s, logvar_], dim=0)

                mu_s, logvar_s = ProductOfExperts()(mu_s, logvar_s)    
                mu_out.append(mu_s)
                logvar_out.append(logvar_s)
                qz_x = hydra.utils.instantiate( 
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_s, logvar=logvar_s
                )
                qz_xs.append(qz_x)
            mu_out = torch.stack(mu_out)
            logvar_out = torch.stack(logvar_out)
            
            moe_mu, moe_logvar = MixtureOfExperts()(mu_out, logvar_out)
            
            qz_x = hydra.utils.instantiate( 
                self.cfg.encoder.default.enc_dist, loc=moe_mu, logvar=moe_logvar
                )
            return [qz_xs, qz_x]
        else:
            for i in range(self.n_views):
                mu_s = mu[i]
                logvar_s = logvar[i] 
                mu_out.append(mu_s)
                logvar_out.append(logvar_s)
              
            mu_ = self.prior.mean
            mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
            logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
            logvar_ = logvar_.expand(logvar[0].shape)
            mu_out.append(mu_)
            logvar_out.append(logvar_)

            mu_out = torch.stack(mu_out)
            logvar_out = torch.stack(logvar_out)

            mu, logvar = ProductOfExperts()(mu_out, logvar_out)
 
            qz_x = hydra.utils.instantiate( 
                self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
                )
            return [qz_x]
        
    def encode_subset(self, x, subset):
        r""" Forward pass through encoder networks for a subset of modalities.
        Args:
            x (list): list of input data of type torch.Tensor.
            subset (list): list of modalities to encode.

        Returns:
            (list): list containing the MoE joint encoding distribution. 
        """
        mu = []
        logvar = []

        for i in subset:
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_) 
            logvar.append(logvar_) 
    
        if len(subset) == self.n_views:
            mu_ = self.prior.mean
            mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
            logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
            logvar_ = logvar_.expand(logvar[0].shape)
            mu.append(mu_)
            logvar.append(logvar_)

        mu = torch.stack(mu)
        logvar = torch.stack(logvar)

        mu, logvar = ProductOfExperts()(mu, logvar)
        qz_x = hydra.utils.instantiate( 
            self.cfg.encoder.default.enc_dist, loc=mu, logvar=logvar
            )
        return [qz_x]
    
    def decode_subset(self, qz_x, subset):
        px_zs = []
        for i in range(self.n_views):
            if i in subset:
                px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
                px_zs.append(px_z)
        return [px_zs]    
    
    def decode(self, qz_x):
        r"""Forward pass of joint latent dimensions through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """    
        px_zs = []
        for i in range(self.n_views):
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate the joint and subset latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        qz_xs, qz_x = self.encode(x)
        px_zs = self.decode([qz_x])
        fwd_rtn = {"px_zs": px_zs, "qz_xs_subsets": qz_xs, "qz_x_joint": qz_x}
        return fwd_rtn

    def loss_function(self, x, fwd_rtn):
        r"""Calculate MoPoE VAE loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the MoPoE VAE loss.
        """
        px_zs = fwd_rtn["px_zs"]
        qz_xs = fwd_rtn["qz_xs_subsets"]

        kl = self.calc_kl_moe(qz_xs)
        ll = self.calc_ll(x, px_zs)

        total = self.beta * kl - ll 

        losses = {"loss": total, "kl": kl, 'll': ll}
        return losses

    def calc_kl_moe(self, qz_xs):
        r"""Calculate KL-divergence between the each PoE subset posterior and the prior distribution.

        Args:
            qz_xs (list): list of encoding distributions.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        weight = 1/len(qz_xs)
        kl = 0
        for qz_x in qz_xs:
            kl +=qz_x.kl_divergence(self.prior).mean(0).sum()

        return kl*weight

    def set_subsets(self, n_views=None):
        """Create combinations of subsets of views.

        Returns:
            subset_list (list): list of unique combinations of n_views.
        """
        if n_views is None:
            n_views = self.n_views
        xs = list(range(0, n_views))
        tmp = [list(combinations(xs, n+1)) for n in range(len(xs))]
        subset_list = [list(item) for sublist in tmp for item in sublist]
        return subset_list

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum()*self.ll_weighting #first index is latent, second index is view
        return ll



import torch
import hydra
from ..base.constants import MODEL_MVAE
from ..base.base_model import BaseModelVAE
from ..base.representations import ProductOfExperts, MeanRepresentation
import numpy as np

class mVAE(BaseModelVAE):
    r"""
    Multimodal Variational Autoencoder (MVAE).

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.beta (int, float): KL divergence weighting term.
            - model.join_type (str): Method of combining encoding distributions.      
            - model.warmup (int): KL term weighted by beta linearly increased to 1 over this many epochs.
            - model.use_prior (bool): Whether to use a prior expert when combining encoding distributions.
            - model.sparse (bool): Whether to enforce sparsity of the encoding distribution.
            - model.threshold (float): Dropout threshold applied to the latent dimensions. Default is 0.
            - model.weight_ll (bool): Whether to weight the log-likelihood loss by 1/n_views.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
        
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Wu, M., & Goodman, N.D. (2018). Multimodal Generative Models for Scalable Weakly-Supervised Learning. NeurIPS.

    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_MVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

        self.join_type = self.cfg.model.join_type
        if self.join_type == "PoE":
            self.join_z = ProductOfExperts()
        elif self.join_type == "Mean":
            self.join_z = MeanRepresentation()
        
        if self.warmup is not None:
            self.beta_vals = torch.linspace(0, self.beta, self.warmup)

        if self.weight_ll:
            self.ll_weighting = 1/self.n_views
        else:
            self.ll_weighting = 1

     #   if self.use_likelihood_rescaling: 
     #       if hasattr(self, 'rescale_factors'):
     #           self.rescale_factors = self.rescale_factors
     #       else:
     #           max_dim = max(*[np.prod(self.input_dim[k]) for k in self.input_dim])
     #           self.rescale_factors = [max_dim / np.prod(self.input_dim[k]) for k in self.input_dim]

     #   else:
     #       self.rescale_factors = [1 for k in self.input_dim]

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): Single element list of joint encoding distribution.
        """
        mu = []
        logvar = []
        for i in range(self.n_views):
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_)
            logvar.append(logvar_)
        if not self.sparse and self.use_prior:
            #get mu and logvar from prior expert
            mu_ = self.prior.mean
            mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
            logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
            logvar_ = logvar_.expand(logvar[0].shape)      
            mu.append(mu_)
            logvar.append(logvar_)
        mu = torch.stack(mu)
        logvar = torch.stack(logvar)
        mu_out, logvar_out = self.join_z(mu, logvar)
    
        qz_x = hydra.utils.instantiate(
            self.cfg.encoder.default.enc_dist, loc=mu_out, logvar=logvar_out
        )
        return [qz_x]

    def encode_subset(self, x, subset):
        r"""Forward pass through encoder networks for the specified subset.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): Single element list of joint encoding distribution.
        """
        mu = []
        logvar = []
        for i in subset:
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_)
            logvar.append(logvar_)
        if not self.sparse and self.use_prior:
            #get mu and logvar from prior expert
            mu_ = self.prior.mean
            mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
            logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
            logvar_ = logvar_.expand(logvar[0].shape)      
            mu.append(mu_)
            logvar.append(logvar_)
        mu = torch.stack(mu)
        logvar = torch.stack(logvar)
        mu_out, logvar_out = self.join_z(mu, logvar)
    
        qz_x = hydra.utils.instantiate(
            self.cfg.encoder.default.enc_dist, loc=mu_out, logvar=logvar_out
        )
        return [qz_x] 
           
    def decode(self, qz_x):
        r"""Forward pass of joint latent dimensions through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """  
        px_zs = []
        for i in range(self.n_views):
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def decode_subset(self, qz_x, subset):
        r"""Forward pass of joint latent dimensions through decoder networks for the specified subset.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element indicating the shared latent dimensions. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """  
        px_zs = []
        for i in subset:
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]
    
    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate the joint latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        qz_x = self.encode(x)
        px_zs = self.decode(qz_x)
        fwd_rtn = {"px_zs": px_zs, "qz_x": qz_x}
        return fwd_rtn

    def calc_kl(self, qz_x):
        r"""Calculate KL-divergence loss.

        Args:
            qz_xs (list): Single element list containing joint encoding distribution.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        if self.sparse:
            kl = qz_x[0].sparse_kl_divergence().mean(0).sum()
        else:
            kl = qz_x[0].kl_divergence(self.prior).mean(0).sum()
        return kl

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum() #*self.rescale_factors[i] #first index is latent, second index is view
        return ll*self.ll_weighting

    def loss_function(self, x, fwd_rtn):
        r"""Calculate Multimodal VAE loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the MVAE loss.
        """
        px_zs = fwd_rtn["px_zs"]
        qz_x = fwd_rtn["qz_x"]

        kl = self.calc_kl(qz_x)
        ll = self.calc_ll(x, px_zs)

        if self.current_epoch > self.warmup -1:
            total = self.beta*kl - ll
        else:
            total = self.beta_vals[self.current_epoch]*kl - ll   
        losses = {"loss": total, "kl": kl, "ll": ll}
        return losses

    def calc_nll(self, x, K=1000, batch_size_K=100):
        r"""Calculate negative log-likelihood used to evaluate model performance.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            nll (torch.Tensor): Negative log-likelihood.
        """
        self._training = False
        ll = 0
        qz_x = self.encode(x)
        zs = qz_x[0].rsample(torch.Size([K]))
        start_idx = 0
        stop_idx = min(start_idx + batch_size_K, K)
        lnpxs = []
        while start_idx < stop_idx:
            zs_ = zs[start_idx:stop_idx]
            lpx_zs = 0
            for j in range(self.n_views):
                px_z = self.decoders[j](zs_)
                lpx_zs += px_z.log_likelihood(x[j]).sum(dim=-1)

            lpz = self.prior.log_likelihood(zs_).sum(dim=-1)
            lqz_xy = qz_x[0].log_likelihood(zs_).sum(dim=-1)
            ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
            lnpxs.append(ln_px)
            start_idx += batch_size_K
            stop_idx = min(stop_idx + batch_size_K, K)
        lnpxs = torch.stack(lnpxs)
        ll += (torch.logsumexp(lnpxs, dim=0) - np.log(K)).mean()
        return -ll


from ..base.constants import MODEL_AE
from ..base.base_model import BaseModelAE

class AE(BaseModelAE):
    r"""Multi-view Autoencoder model with a separate latent representation for each view.

    Args:
        cfg (str): Path to configuration file.
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.
    """
    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_AE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            z (list): list of latent dimensions for each view of type torch.Tensor.
        """   
        z = []
        for i in range(self.n_views):
            z_ = self.encoders[i](x[i])
            z.append(z_)
        return z

    def decode(self, z):
        r"""Forward pass through decoder networks. Each latent is passed through all of the decoders.

        Args:
            z (list): list of latent dimensions for each view of type torch.Tensor.

        Returns:
            x_recon (list): list of data reconstructions.
        """
        x_recon = []
        for i in range(self.n_views):
            temp_recon = [self.decoders[j](z[i]) for j in range(self.n_views)]
            x_recon.append(temp_recon)
        return x_recon

    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing list of data reconstructions (x_recon) and latent dimensions (z).
        """
        z = self.encode(x)
        x_recon = self.decode(z)
        fwd_rtn = {"x_recon": x_recon, "z": z}
        return fwd_rtn

    def loss_function(self, x, fwd_rtn):    
        r"""Calculate reconstruction loss.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing list of data reconstructions (x_recon) and latent dimensions (z).

        Returns:
            losses (dict): dictionary containing reconstruction loss.
        """
        x_recon = fwd_rtn["x_recon"]
        recon = 0
        for i in range(self.n_views):
            for j in range(self.n_views):
                recon += - x_recon[j][i].log_likelihood(x[i]).mean(0).sum() #first index is latent, second index is view
        recon = recon / self.n_views / self.n_views
        losses = {"loss": recon}
        return losses



from schema import Schema, And, Or, Optional, SchemaError, Regex

SUPPORTED_ENCODERS = [
            "mlp.Encoder",
            "mlp.VariationalEncoder",
            "mlp.ConditionalVariationalEncoder",
            "cnn.Encoder",
            "cnn.VariationalEncoder"
        ]

SUPPORTED_DECODERS = [
            "mlp.Decoder",
            "mlp.VariationalDecoder",
            "mlp.ConditionalVariationalDecoder",
            "cnn.Decoder"
        ]

SUPPORTED_DISCRIMINATORS = [
            "mlp.Discriminator"
        ]

SUPPORTED_DISTRIBUTIONS = [
            "distributions.Default",
            "distributions.Categorical",
            "distributions.Normal",
            "distributions.MultivariateNormal",
            "distributions.Bernoulli",
            "distributions.ApproxBernoulli",
            "distributions.Laplace"
        ]

SUPPORTED_JOIN = [
            "PoE",
            "Mean"
        ]

SUPPORTED_DATASETS = ["multiviewae.base.datasets.MVDataset",
                      "multiviewae.base.datasets.IndexMVDataset",
                ]

UNSUPPORTED_ENC_DIST = [
            "distributions.Bernoulli"
        ]

UNSUPPORTED_PRIOR_DIST = [
            "distributions.Default",
            "distributions.Bernoulli"
        ]

def return_or(params=[], msg="invalid"):
    assert(len(params) > 0)

    p = ', '.join('"{0}"'.format(str(w)) for w in params)
    return f"Or({p}, error='{msg}')"

def return_regexor(params=[], msg="invalid"):
    assert(len(params) > 0)
    p = '|'.join('{0}'.format(str(w)) for w in params)
    return f"Regex(r'(.*?).({p})$', error='{msg}')"

def list_sub(a, b):
    return list(set(a) - set(b))

config_schema = Schema({
    "model": {
        "save_model": bool,
        "seed_everything": bool,
        "seed": And(int, lambda x: 0 <= x <= 4294967295),
        "z_dim": int,
        "learning_rate": And(float, lambda x: 0 < x < 1),
        "sparse": bool,
        "threshold": Or(And(float, lambda x: 0 < x < 1), 0),
        Optional("eps"): And(float, lambda x: 0 < x <= 1e-10),
        Optional("beta"): And(Or(int, float), lambda x: x > 0),
        Optional("K"): And(int, lambda x: x >= 1),
        Optional("alpha"): And(Or(int, float), lambda x: x > 0),
        Optional("private"): bool,
        Optional("join_type"): eval(return_or(params=SUPPORTED_JOIN,
                        msg="model.join_type: unsupported or invalid join type"))
    },
    "datamodule": {
        "_target_": Regex(r'(.*?)DataModule$'),   
        "batch_size": Or(And(int, lambda x: x > 0), None),
        "is_validate": bool,
        "train_size": And(float, lambda x: 0 < x < 1),
        "dataset": {
            "_target_" : Regex(r'(.*?)Dataset$'),
            Optional("data_dir"): str, 
            Optional("views"): [And(int, lambda x: x >= 0)],}
    },
    "encoder": {
        "default": {
            "_target_" : eval(return_regexor(params=SUPPORTED_ENCODERS,
                            msg="encoder._target_: unsupported or invalid encoder")),
            Optional("hidden_layer_dim"): [And(int, lambda x: x > 0)],
            Optional(Regex(r'^layer\d$')) : {
                "layer": str, 
            },
            Optional("bias"): bool,
            Optional("non_linear"): bool,
            Optional("num_cat"): And(int, lambda x: x > 1),
            "enc_dist": {
                    "_target_": eval(return_regexor(params=list_sub(SUPPORTED_DISTRIBUTIONS, UNSUPPORTED_ENC_DIST),
                            msg="encoder.enc_dist._target_: unsupported or invalid encoder distribution"))
            }
        },
        Optional(Regex(r'^enc\d$')) : {
            "_target_" : eval(return_regexor(params=SUPPORTED_ENCODERS,
                            msg="encoder._target_: unsupported or invalid encoder")),
            Optional("hidden_layer_dim"): [And(int, lambda x: x > 0)],
            Optional(Regex(r'^layer\d$')) : {
                "layer": str, 
            },
            Optional("bias"): bool,
            Optional("non_linear"): bool,
            Optional("num_cat"): And(int, lambda x: x > 1),
            "enc_dist": {
                    "_target_": eval(return_regexor(params=list_sub(SUPPORTED_DISTRIBUTIONS, UNSUPPORTED_ENC_DIST),
                            msg="encoder.enc_dist._target_: unsupported or invalid encoder distribution"))
            }
        }
    },
    "decoder": {
        "default": {
            "_target_" : eval(return_regexor(params=SUPPORTED_DECODERS,
                            msg="decoder._target_: unsupported or invalid decoder")),
            Optional("hidden_layer_dim"): [And(int, lambda x: x > 0)],
            Optional(Regex(r'^layer\d$')) : {
                "layer": str, 
            },
            Optional("bias"): bool,
            Optional("non_linear"): bool,
            Optional("num_cat"): And(int, lambda x: x > 1),
            Optional("init_logvar"): Or(int, float),
            "dec_dist": {
                    "_target_": eval(return_regexor(params=SUPPORTED_DISTRIBUTIONS,
                            msg="decoder.dec_dist._target_: unsupported or invalid decoder distribution"))
            }
        },
        Optional(Regex(r'^dec\d$')) : {
            "_target_" : eval(return_regexor(params=SUPPORTED_DECODERS,
                            msg="decoder._target_: unsupported or invalid decoder")),
            Optional("hidden_layer_dim"): [And(int, lambda x: x > 0)],
            Optional(Regex(r'^layer\d$')) : {
                "layer": str, 
            },
            Optional("bias"): bool,
            Optional("non_linear"): bool,
            Optional("num_cat"): And(int, lambda x: x > 1),
            Optional("init_logvar"): Or(int, float),
            "dec_dist": {
                    "_target_": eval(return_regexor(params=SUPPORTED_DISTRIBUTIONS,
                            msg="decoder.dec_dist._target_: unsupported or invalid decoder distribution"))
            }
        }
    },
    Optional("discriminator"): {
        "_target_" : eval(return_regexor(params=SUPPORTED_DISCRIMINATORS,
                        msg="discriminator._target_: unsupported or invalid discriminator")),
        "hidden_layer_dim": [And(int, lambda x: x > 0)],
        Optional("bias"): bool,
        Optional("non_linear"): bool,
        "dropout_threshold": Or(0, And(float, lambda x: 0 < x < 1))
    },
    "prior": {
       "_target_" : eval(return_regexor(params=list_sub(SUPPORTED_DISTRIBUTIONS, UNSUPPORTED_PRIOR_DIST),
                        msg="prior._target_: unsupported or invalid prior")),
       "loc":  Or(float, [float]),
       "scale": Or(And(float, lambda x: x > 0), [And(float, lambda x: x > 0)])
    },
    "trainer": {
       "_target_" : "pytorch_lightning.Trainer",
       "accelerator": Or("cpu", "gpu", "auto"),
       "max_epochs": And(int, lambda x: x > 0),
       "deterministic": bool,
       "log_every_n_steps": And(int, lambda x: x > 0)
    },
    "callbacks": {
        "model_checkpoint": {   
           "_target_" : "pytorch_lightning.callbacks.ModelCheckpoint",
           "monitor": Or("train_loss", "val_loss"), # see training_step() and validation_step()
           "mode": Or("min", "max"),
           "save_last": bool,
           "dirpath": str   
        },
        "early_stopping": { 
           "_target_" : "pytorch_lightning.callbacks.EarlyStopping",
           "monitor": Or("train_loss", "val_loss"), # see training_step() and validation_step()
           "mode": Or("min", "max"),
           "patience": And(int, lambda x: x > 0),
           "min_delta": float,
           "verbose": bool
        }
    },
    "logger": {
       "_target_" : str,
       "save_dir": str 
    }
}, ignore_extra_keys=True)



import torch
import torch.nn as nn
from .constants import EPS

class ProductOfExperts(nn.Module):
    """Return parameters for product of independent experts.

    Args:
    mu (torch.Tensor): Mean of experts distribution. M x D for M experts
    logvar (torch.Tensor): Log of variance of experts distribution. M x D for M experts
    """
    def forward(self, mu, logvar):
        var = torch.exp(logvar) + EPS
        # precision of i-th Gaussian expert at point x
        T = 1.0 / var
        pd_mu = torch.sum(mu * T, dim=0) / torch.sum(T, dim=0)
        pd_var = 1.0 / torch.sum(T, dim=0)
        pd_logvar = torch.log(pd_var)
        return pd_mu, pd_logvar
    
class alphaProductOfExperts(nn.Module):
    """Return parameters for weighted product of independent experts (mmJSD implementation).
    See https://arxiv.org/pdf/1410.7827.pdf for equations.

    Args:
    mu (torch.Tensor): Mean of experts distribution. M x D for M experts
    logvar (torch.Tensor): Log of variance of experts distribution. M x D for M experts
    """

    def forward(self, mu, logvar, weights=None):
        if weights is None:
            num_components = mu.shape[0]
            weights = (1/num_components) * torch.ones(mu.shape).to(mu.device)

        var = torch.exp(logvar) + EPS
        T = 1. / (var + EPS)
        weights = torch.broadcast_to(weights, mu.shape)
        pd_var = 1. / torch.sum(weights * T + EPS, dim=0)
        pd_mu = pd_var * torch.sum(weights * mu * T, dim=0)
        pd_logvar = torch.log(pd_var + EPS)
        return pd_mu, pd_logvar
    
class weightedProductOfExperts(nn.Module):
    """Return parameters for weighted product of independent experts.
    See https://arxiv.org/pdf/1410.7827.pdf for equations.

    Args:
    mu (torch.Tensor): Mean of experts distribution. M x D for M experts
    logvar (torch.Tensor): Log of variance of experts distribution. M x D for M experts
    """

    def forward(self, mu, logvar, weight):

        var = torch.exp(logvar) + EPS     
        weight = weight[:, None, :].repeat(1, mu.shape[1],1)
        T = 1.0 / (var + EPS)
        pd_var = 1. / torch.sum(weight * T + EPS, dim=0)
        pd_mu = pd_var * torch.sum(weight * mu * T, dim=0)
        pd_logvar = torch.log(pd_var + EPS)
        return pd_mu, pd_logvar

class MixtureOfExperts(nn.Module):
    """Return parameters for mixture of independent experts.
    Implementation from: https://github.com/thomassutter/MoPoE

    Args:
    mus (torch.Tensor): Mean of experts distribution. M x D for M experts
    logvars (torch.Tensor): Log of variance of experts distribution. M x D for M experts
    """

    def forward(self, mus, logvars):

        num_components = mus.shape[0]
        num_samples = mus.shape[1]
        weights = (1/num_components) * torch.ones(num_components).to(mus[0].device)
        idx_start = []
        idx_end = []
        for k in range(0, num_components):
            if k == 0:
                i_start = 0
            else:
                i_start = int(idx_end[k-1])
            if k == num_components-1:
                i_end = num_samples
            else:
                i_end = i_start + int(torch.floor(num_samples*weights[k]))
            idx_start.append(i_start)
            idx_end.append(i_end)
        idx_end[-1] = num_samples

        mu_sel = torch.cat([mus[k, idx_start[k]:idx_end[k], :] for k in range(num_components)])
        logvar_sel = torch.cat([logvars[k, idx_start[k]:idx_end[k], :] for k in range(num_components)])

        return mu_sel, logvar_sel

class MeanRepresentation(nn.Module):
    """Return mean of separate VAE representations.
    
    Args:
    mu (torch.Tensor): Mean of distributions. M x D for M views.
    logvar (torch.Tensor): Log of Variance of distributions. M x D for M views.
    """

    def forward(self, mu, logvar, **kwargs):
        mean_mu = torch.mean(mu, axis=0)
        mean_logvar = torch.mean(logvar, axis=0)
        
        return mean_mu, mean_logvar



import math
import torch
import hydra

from ..base.constants import MODEL_MMVAE
from ..base.base_model import BaseModelVAE
import numpy as np
#set numpy seed
np.random.seed(0)

class mmVAE(BaseModelVAE):
    r"""
    Mixture-of-Experts Multimodal Variational Autoencoder (MMVAE). 

    Code is based on: https://github.com/iffsid/mmvae

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.K (int): Number of samples to take from encoding distribution.
            - model.DREG_loss (bool): Whether to use DReG estimator when using large K value. 
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar (int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
            
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions. 
    
    References
    ----------
    Shi, Y., Siddharth, N., Paige, B., & Torr, P.H. (2019). Variational Mixture-of-Experts Autoencoders for Multi-Modal Deep Generative Models. ArXiv, abs/1911.03393.
    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_MMVAE,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

    def encode(self, x):
        r"""Forward pass through encoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): list of encoding distributions.
        """
        qz_xs = []
        for i in range(self.n_views):
            mu, logvar = self.encoders[i](x[i])
            qz_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu, logvar=logvar
            )
            qz_xs.append(qz_x)
        return qz_xs
    
    def encode_subset(self, x, subset):
        r"""Forward pass through encoder networks for a subset of modalities.
        Args:
            x (list): list of input data of type torch.Tensor.
            subset (list): list of modalities to encode.

        Returns:
            (list): list of encoding distributions.
        """    
        qz_xs = []
        for i in range(self.n_views):
            if i in subset:
                mu, logvar = self.encoders[i](x[i])
                qz_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu, logvar=logvar
                )
            else:
            # Choose one of the subset modalities at random
                mod = np.random.choice(subset)
                mu, logvar = self.encoders[mod](x[mod])
                qz_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{mod}.enc_dist"), loc=mu, logvar=logvar
                )
            qz_xs.append(qz_x)
        return qz_xs
        
    def decode(self, qz_xs):
        r"""Forward pass through decoder networks. Each latent is passed through all of the decoders.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions. The outer list has a n_view element indicating latent dimensions index. 
            The inner list is a n_view element list with the position in the list indicating the decoder index.
        """    
        px_zs = []
        for qz_x in qz_xs:
            if self._training:
                px_z = [
                    self.decoders[j](qz_x.rsample(torch.Size([self.K])))
                    for j in range(self.n_views)
                ]
            else:
                px_z = [
                    self.decoders[j](qz_x._sample())
                    for j in range(self.n_views)
                ]
            px_zs.append(
                px_z
            )  
            del px_z
        return px_zs

    def decode_subset(self, qz_xs, subset):
        r"""Forward pass through decoder networks for a subset of modalities. Each latent is passed through its own decoder.

        Args:
            qz_xs (list): list of encoding distributions.
            subset (list): list of modalities to decode.

        Returns:
            (list): A nested list of decoding distributions. The outer list is a single element list, the inner list is a subset element list of decoding distributions.
        """    
        px_zs = []
        for i, qz_x in enumerate(qz_xs):
            if i in subset:
                px_z = self.decoders[i](qz_x._sample(training=self._training, return_mean=self.return_mean))
                px_zs.append(px_z)
        return [px_zs]
            
    
    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding (qz_xs) and decoding (px_zs) distributions.
        """
        qz_xs = self.encode(x)
        px_zs = self.decode(qz_xs)
        return {"qz_xs": qz_xs, "px_zs": px_zs}

    def loss_function(self, x, fwd_rtn):
        r"""Wrapper function for mmVAE loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing mmVAE loss.
        """
        qz_xs, px_zs = fwd_rtn["qz_xs"], fwd_rtn["px_zs"]
        total = -self.moe_iwae(x, qz_xs, px_zs)
        losses = {"loss": total}
        return losses

    def moe_iwae(self, x, qz_xs, px_zs):
        r"""Calculate Mixture-of-Experts importance weighted autoencoder (IWAE) loss used for the mmVAE model.
        
        Args:
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            (torch.Tensor): the output tensor.
        """
        lws = []
        zss = []
        if self.DREG_loss: 
            qz_xs = [hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=qz_x.loc.detach(), logvar=qz_x.logvar.detach()) for i, qz_x in enumerate(qz_xs)]
        for i in range(self.n_views):
            if self._training:
                zs = qz_xs[i].rsample(torch.Size([self.K]))
            else:
                zs = qz_xs[i]._sample()
            zss.append(zs)

        for r, qz_x in enumerate(qz_xs): 
            lpz = self.prior.log_likelihood(zss[r]).sum(-1)
            lqz_x = self.log_mean_exp(
                torch.stack([qz_x.log_likelihood(zss[r]).sum(-1) for qz_x in qz_xs])
            )  # summing over M modalities for each z to create q(z|x1:M)
            
            lpx_z = [
                px_z.log_likelihood(x[d]).view(*px_z._sample().size()[:2], -1).sum(-1)
                for d, px_z in enumerate(px_zs[r])
            ]  # summing over each decoder
            lpx_z = torch.stack(lpx_z).sum(0)

            lw = lpx_z + self.beta*(lpz - lqz_x)
            lws.append(lw)
        if self.DREG_loss:
            zss = torch.stack(zss)
            with torch.no_grad():
                grad_wt = (lws - torch.logsumexp(lws, 1, keepdim=True)).exp()
                if zss.requires_grad:
                    zss.register_hook(lambda grad: grad_wt.unsqueeze(-1) * grad)
            return (grad_wt * lws).sum(0) / self.n_views #DReG loss

        return (
            self.log_mean_exp(torch.stack(lws), dim=1).mean(0).sum()/self.n_views
          )  # looser iwae bound 
            

    def log_mean_exp(self, value, dim=0, keepdim=False):
        r"""Returns the log of the mean of the exponentials along the given dimension (dim).

        Args:
            value (torch.Tensor): the input tensor.
            dim (int, optional): the dimension along which to take the mean.
            keepdim (bool, optional): whether the output tensor has dim retained or not.

        Returns:
            (torch.Tensor): the output tensor.
        """
        return torch.logsumexp(value, dim, keepdim=keepdim) - math.log(value.size(dim))



import torch
import hydra

from ..base.constants import MODEL_MMJSD
from ..base.base_model import BaseModelVAE
from ..base.representations import MixtureOfExperts, alphaProductOfExperts

class mmJSD(BaseModelVAE):
    r"""
    Multimodal Jensen-Shannon divergence (mmJSD) model with Product-of-Experts dynamic prior. 

    Code is based on: https://github.com/thomassutter/mmjsd

    Args:
        cfg (str): Path to configuration file. Model specific parameters in addition to default parameters:
            
            - model.private (bool): Whether to include private modality-specific latent dimensions.
            - model.beta (int, float): KL divergence weighting term.
            - model.alpha (int, float): JSD divergence weighting term.
            - model.s_dim (int): Number of private latent dimensions.
            - encoder.default._target_ (multiviewae.architectures.mlp.VariationalEncoder): Type of encoder class to use.
            - encoder.default.enc_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Encoding distribution.
            - decoder.default._target_ (multiviewae.architectures.mlp.VariationalDecoder): Type of decoder class to use.
            - decoder.default.init_logvar(int, float): Initial value for log variance of decoder.
            - decoder.default.dec_dist._target_ (multiviewae.base.distributions.Normal, multiviewae.base.distributions.MultivariateNormal): Decoding distribution.
    
        input_dim (list): Dimensionality of the input data.
        z_dim (int): Number of latent dimensions.

    References
    ----------
    Sutter, Thomas & Daunhawer, Imant & Vogt, Julia. (2021). Multimodal Generative Learning Utilizing Jensen-Shannon-Divergence. Advances in Neural Information Processing Systems. 33. 

    """

    def __init__(
        self,
        cfg = None,
        input_dim = None,
        z_dim = None
    ):
        super().__init__(model_name=MODEL_MMJSD,
                        cfg=cfg,
                        input_dim=input_dim,
                        z_dim=z_dim)

        if self.weight_ll:
            self.ll_weighting = 1/self.n_views
        else:
            self.ll_weighting = 1
            
    def encode(self, x):
        r"""Forward pass through encoder network. If self.private=True, the first two dimensions of each latent are used for the modality-specific part and the remaining dimensions for the joint content.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            Returns a combination of the following depending on the training stage and model type: 
            qz_xs (list): Single element list containing the MoE encoding distribution for self.private=False.
            qzs_xs (list): list containing each encoding distribution for self.private=False.
            qz_x (list):  Single element list containing the PoE encoding distribution for self.private=False.

            qc_x (list): Single element list containing the PoE shared encoding distribution.
            qscs_xs (list): list containing combined shared and private latents.
            qs_xs (list): list of encoding distributions for private latents.
            qcs_xs (list): list containing encoding distributions for shared latent dimensions for each view.
        """
     
        if self.private:
            qs_xs = []
            qcs_xs = []
            mu_s = []
            logvar_s = []
            mu_c = []
            logvar_c = []
            for i in range(self.n_views):
                mu, logvar = self.encoders[i](x[i])
                mu_s.append(mu[:,:self.s_dim])
                logvar_s.append(logvar[:,:self.s_dim])
                mu_c.append(mu[:,self.s_dim:])
                logvar_c.append(logvar[:,self.s_dim:])

                qs_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu[:,:self.s_dim], logvar=logvar[:,:self.s_dim]
                )
                qs_xs.append(qs_x)
                qc_x = hydra.utils.instantiate(
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu[:,self.s_dim:], logvar=logvar[:,self.s_dim:]
                )
                qcs_xs.append(qc_x)

            mu_c = torch.stack(mu_c)
            logvar_c = torch.stack(logvar_c)

            moe_mu_c, moe_logvar_c = MixtureOfExperts()(mu_c, logvar_c)
       
            poe_mu_c, poe_logvar_c = alphaProductOfExperts()(mu_c, logvar_c)
            qc_x = hydra.utils.instantiate( 
            self.cfg.encoder.default.enc_dist, loc=poe_mu_c, logvar=poe_logvar_c
            )
            qscs_xs = []
            for i in range(self.n_views):       
                mu_sc = torch.cat((mu_s[i], moe_mu_c), 1)
                logvar_sc = torch.cat((logvar_s[i], moe_logvar_c), 1)
                qsc_x = hydra.utils.instantiate( 
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_sc, logvar=logvar_sc
                )
                qscs_xs.append(qsc_x)

            if self._training:
                return [[qc_x], qcs_xs, qs_xs, qscs_xs]
            
            qscs_xs = []
            for i in range(self.n_views):       
                mu_sc = torch.cat((mu_s[i], poe_mu_c), 1)
                logvar_sc = torch.cat((logvar_s[i], poe_logvar_c), 1)
                qsc_x = hydra.utils.instantiate( 
                eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_sc, logvar=logvar_sc
                )
                qscs_xs.append(qsc_x)            
            return qscs_xs   

        mu = []
        logvar = []
        qzs_xs = []
        for i in range(self.n_views):
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_) 
            logvar.append(logvar_) 
            qz_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_, logvar=logvar_
            )
            qzs_xs.append(qz_x)

        #add prior expert
        mu_ = self.prior.mean
        mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
        logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
        logvar_ = logvar_.expand(logvar[0].shape)
        qz_x = hydra.utils.instantiate(
            eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_, logvar=logvar_
            )
        mu.append(mu_)
        logvar.append(logvar_)
        qzs_xs.append(qz_x)

        mu = torch.stack(mu)
        logvar = torch.stack(logvar)

        moe_mu, moe_logvar = MixtureOfExperts()(mu, logvar)

        qz_xs =  hydra.utils.instantiate(
                self.cfg.encoder.default.enc_dist, loc=moe_mu, logvar=moe_logvar
            )
        
        poe_mu, poe_logvar = alphaProductOfExperts()(mu, logvar)
        qz_x = hydra.utils.instantiate( 
                self.cfg.encoder.default.enc_dist, loc=poe_mu, logvar=poe_logvar
            )
        if self._training:
            return [[qz_xs], qzs_xs, [qz_x]]

        return [qz_x]
    
    def encode_subset(self, x, subset):
        r""" Forward pass through encoder networks for a subset of modalities.
        Args:
            x (list): list of input data of type torch.Tensor.
            subset (list): list of modalities to encode.

        Returns:
            (list): list containing the PoE joint encoding distribution. 
        """
        assert (self.private==False), \
        "Subset feature only works for private=False (for now)." 
        mu = []
        logvar = []
        qzs_xs = []
        for i in subset:
            mu_, logvar_ = self.encoders[i](x[i])
            mu.append(mu_) 
            logvar.append(logvar_) 
            qz_x = hydra.utils.instantiate(
                    eval(f"self.cfg.encoder.enc{i}.enc_dist"), loc=mu_, logvar=logvar_
            )
            qzs_xs.append(qz_x)
        if len(subset) > 1:
            mu_ = self.prior.mean
            mu_ = mu_.expand(mu[0].shape).to(mu[0].device)           
            logvar_ = torch.log(self.prior.variance).to(mu[0].device)     
            logvar_ = logvar_.expand(logvar[0].shape)
            mu.append(mu_)
            logvar.append(logvar_)

        mu = torch.stack(mu)
        logvar = torch.stack(logvar)

        poe_mu, poe_logvar = alphaProductOfExperts()(mu, logvar)

        qz_x =  hydra.utils.instantiate(
                self.cfg.encoder.default.enc_dist, loc=poe_mu, logvar=poe_logvar
            )

        return [qz_x]
    
    def decode(self, qz_x):
        r"""Forward pass of latent dimensions through decoder networks.

        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element, the inner list is a n_view element list with the position in the list indicating the decoder index.
        """      
        px_zs = []
        for i in range(self.n_views):
            if self.private:
                px_z = self.decoders[i](qz_x[i]._sample(training=self._training, return_mean=self.return_mean))
            else:
                px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]

    def decode_subset(self, qz_x, subset):
        r"""Forward pass of latent dimensions through decoder networks for a subset of modalities.

        Args:
            x (list): list of input data of type torch.Tensor.
            subset (list): list of modalities to decode.

        Returns:
            (list): A nested list of decoding distributions, px_zs. The outer list has a single element, the inner list is a n_view element list with the position in the list indicating the decoder index.
        """      
        assert (self.private==False), \
        "Subset feature only works for private=False (for now)." 
        px_zs = []
        for i in subset:
            px_z = self.decoders[i](qz_x[0]._sample(training=self._training, return_mean=self.return_mean))
            px_zs.append(px_z)
        return [px_zs]
    
    def forward(self, x):
        r"""Apply encode and decode methods to input data to generate the joint and modality specific latent dimensions and data reconstructions. 
        
        Args:
            x (list): list of input data of type torch.Tensor.

        Returns:
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.
        """
        if self.private:
            qc_x, qcs_xs, qs_xs, qscs_xs = self.encode(x)
            px_zs = self.decode(qscs_xs)
            fwd_rtn = {"px_zs": px_zs, "qcs_xs": qcs_xs, "qs_xs": qs_xs, "qc_x": qc_x}
        else:
            qz_xs, qzs_xs, qz_x = self.encode(x)
            px_zs = self.decode(qz_xs)
            fwd_rtn = {"px_zs": px_zs, "qzs_xs": qzs_xs, "qz_x": qz_x}
        return fwd_rtn
        
    def calc_kl(self, qz_xs):
        r"""Calculate KL-divergence loss.

        Args:
            qz_xs (list): list of encoding distributions.

        Returns:
            (torch.Tensor): KL-divergence loss.
        """
        kl = 0
        for i in range(len(qz_xs)):
            kl += qz_xs[i].kl_divergence(self.prior).sum(1, keepdims=True).mean(0)
        return self.beta*kl/self.n_views

    def calc_ll(self, x, px_zs):
        r"""Calculate log-likelihood loss.

        Args:
            x (list): list of input data of type torch.Tensor.
            px_zs (list): list of decoding distributions.

        Returns:
            ll (torch.Tensor): Log-likelihood loss.
        """
        ll = 0
        for i in range(self.n_views):
            ll += px_zs[0][i].log_likelihood(x[i]).mean(0).sum()*self.ll_weighting #first index is latent, second index is view 
        return ll
    
    def calc_jsd(self, qcs_xs, qc_x):
        r"""Calculate Jensen-Shannon Divergence loss.

        Args:
            qcs_xs (list): list of encoding distributions of each view for shared latent dimensions.
            qc_x (list): Dynamic prior given by PoE of shared encoding distributions.

        Returns:
            jsd (torch.Tensor): Jensen-Shannon Divergence loss.
        """
        jsd = 0
        for i in range(self.n_views):
            jsd += qcs_xs[i].kl_divergence(qc_x[0]).sum(1, keepdims=True).mean(0)
        return self.alpha*jsd/(self.n_views+1)

    def loss_function(self, x, fwd_rtn):
        r"""Calculate mmJSD loss.
        
        Args:                                                                                                                           
            x (list): list of input data of type torch.Tensor.
            fwd_rtn (dict): dictionary containing encoding and decoding distributions.

        Returns:
            losses (dict): dictionary containing each element of the mmJSD loss.
        """
        px_zs = fwd_rtn["px_zs"]
        ll = self.calc_ll(x, px_zs)

        if self.private:
            qcs_xs = fwd_rtn["qcs_xs"]
            qs_xs = fwd_rtn["qs_xs"] 
            qc_x = fwd_rtn["qc_x"]
            jsd = self.calc_jsd(qcs_xs, qc_x)
            kl = self.calc_kl(qs_xs)
            total = kl + jsd - ll
            losses = {"loss": total, "kl": kl, "ll": ll, "jsd": jsd}
        else:
            qz_x = fwd_rtn["qz_x"]
            qzs_xs = fwd_rtn["qzs_xs"]
            jsd = self.calc_jsd(qzs_xs, qz_x)
            total = jsd - ll
            losses = {"loss": total, "ll": ll, "jsd": jsd}
        return losses
